@@ -17,9 +17,12 @@
 /* GeoTIFF overrides */
 
 #include "geotiff.h"
+#include "geo_normalize.h"
 #include "geo_tiffp.h"
 #include "geo_keyp.h"
 #include "xtiffio.h"
+#include "cpl_serv.h"
+
 #define TIFFOpen XTIFFOpen
 #define TIFFClose XTIFFClose
 
@@ -53,6 +56,7 @@ static	int jpegcolormode = JPEGCOLORMODE_RGB;
 static	uint16 defcompression = (uint16) -1;
 static	uint16 defpredictor = (uint16) -1;
 static 	char *geofile=(char *)0;
+static  char *proj4_string = (char *) 0;
 static  char *worldfile=(char *)0;
 
 static  void ApplyWorldFile(const char *worldfile, TIFF *out);
@@ -72,15 +76,18 @@ main(int argc, char* argv[])
 	uint32 diroff = 0;
 	TIFF* in;
 	TIFF* out;
-	const char* mode = "w";
+	char mode[10];
+	char* mp = mode;
 	int c;
 	extern int optind;
 	extern char* optarg;
 
-	while ((c = getopt(argc, argv, "c:f:l:o:p:r:w:e:g:aistd")) != -1)
+	*mp++ = 'w';
+	*mp = '\0';
+	while ((c = getopt(argc, argv, "c:f:l:o:p:r:w:e:g:4:aistd8BLMC")) != -1)
 		switch (c) {
 		case 'a':		/* append to output */
-			mode = "a";
+			mode[0] = 'a';
 			break;
 		case 'd':		/* down cast 8bit to 4bit */
                         convert_8_to_4 = 1;
@@ -106,6 +113,9 @@ main(int argc, char* argv[])
 		case 'g':		/* GeoTIFF metadata file */
 			geofile = optarg;
 			break;
+		case '4':	       
+			proj4_string = optarg;
+			break;
 		case 'l':		/* tile length */
 			outtiled = TRUE;
 			deftilelength = atoi(optarg);
@@ -134,12 +144,28 @@ main(int argc, char* argv[])
 			outtiled = TRUE;
 			deftilewidth = atoi(optarg);
 			break;
+		case 'B':
+			*mp++ = 'b'; *mp = '\0';
+			break;
+		case 'L':
+			*mp++ = 'l'; *mp = '\0';
+			break;
+		case 'M':
+			*mp++ = 'm'; *mp = '\0';
+			break;
+		case 'C':
+			*mp++ = 'c'; *mp = '\0';
+			break;
+		case '8':
+			*mp++ = '8'; *mp = '\0';
+			break;
 		case '?':
 			usage();
 			/*NOTREACHED*/
 		}
 	if (argc - optind < 2)
 		usage();
+        printf( "mode=%s\n", mode);
 	out = TIFFOpen(argv[argc-1], mode);
 	if (out == NULL)
 		return (-2);
@@ -178,6 +204,7 @@ static void ApplyWorldFile(const char *worldfilename, TIFF *out)
 {
     FILE	*tfw;
     double	pixsize[3], xoff, yoff, tiepoint[6], x_rot, y_rot;
+    int         success;
 
     /* 
      * Read the world file.  Note we currently ignore rotational coefficients!
@@ -189,15 +216,21 @@ static void ApplyWorldFile(const char *worldfilename, TIFF *out)
         return;
     }
 
-    fscanf( tfw, "%lf", pixsize + 0 );
-    fscanf( tfw, "%lf", &y_rot );
-    fscanf( tfw, "%lf", &x_rot );
-    fscanf( tfw, "%lf", pixsize + 1 );
-    fscanf( tfw, "%lf", &xoff );
-    fscanf( tfw, "%lf", &yoff );
+    success  = fscanf( tfw, "%lf", pixsize + 0 );
+    success &= fscanf( tfw, "%lf", &y_rot );
+    success &= fscanf( tfw, "%lf", &x_rot );
+    success &= fscanf( tfw, "%lf", pixsize + 1 );
+    success &= fscanf( tfw, "%lf", &xoff );
+    success &= fscanf( tfw, "%lf", &yoff );
 
     fclose( tfw );
 
+    if( success != 1 )
+    {
+        fprintf( stderr, "Failure parsing one or more lines of world file.\n");
+        return;
+    }
+
     /*
      * Write out pixel scale, and tiepoint information.
      */
@@ -235,7 +268,7 @@ static void ApplyWorldFile(const char *worldfilename, TIFF *out)
 
 static void InstallGeoTIFF(TIFF *out)
 {
-    GTIF *gtif=(GTIF*)0; /* GeoKey-level descriptor */
+    GTIF *gtif; /* GeoKey-level descriptor */
     FILE *fd;
 
     gtif = GTIFNew(out);
@@ -245,21 +278,38 @@ static void InstallGeoTIFF(TIFF *out)
         return;
     }
 
-    /* Install keys and tags */
-    fd = fopen(geofile,"r");
-    if (!GTIFImport(gtif,0,fd)) goto bad;
-    fclose(fd);
+    if( geofile )
+    {
+        /* Install keys and tags */
+        fd = fopen(geofile,"r");
+        if( fd == NULL )
+        {
+            perror( geofile );
+            exit( -1 );
+        }
+        if (!GTIFImport(gtif,0,fd))
+        {
+            fprintf(stderr,"Failure in GTIFImport\n");
+            exit (-1);
+        }
+        fclose(fd);
+    }
+    else if( proj4_string )
+    {
+        if( !GTIFSetFromProj4(gtif,proj4_string) )
+        {
+            fprintf(stderr,"Failure in GTIFSetFromProj4\n");
+            exit (-1);
+        }
+    }
     GTIFWriteKeys(gtif);
     GTIFFree(gtif);
     return;
-  bad:
-    fprintf(stderr,"Failure in GTIFImport\n");
-    exit (-1);
 }
 
 static void CopyGeoTIFF(TIFF * in, TIFF *out)
 {
-    GTIF *gtif=(GTIF*)0; /* GeoKey-level descriptor */
+    GTIF *gtif; /* GeoKey-level descriptor */
     double *d_list = NULL;
     int16   d_list_count;
 
@@ -345,8 +395,10 @@ char* stuff[] = {
 "usage: gtiffcp [options] input... output",
 "where options are:",
 " -g file	install GeoTIFF metadata from <file>",
+" -4 proj4_str	install GeoTIFF metadata from proj4 string",
 " -e file	install positioning info from ESRI Worldfile <file>",
 " -a		append to output instead of overwriting",
+" -8		write BigTIFF instead of default ClassicTIFF",
 " -o offset	set initial directory offset",
 " -p contig	pack samples contiguously (e.g. RGBRGB...)",
 " -p separate	store samples separately (e.g. RRR...GGG...BBB...)",
@@ -452,7 +504,7 @@ static struct cpTag {
 	{ TIFFTAG_DATETIME,		1, TIFF_ASCII },
 	{ TIFFTAG_ARTIST,		1, TIFF_ASCII },
 	{ TIFFTAG_HOSTCOMPUTER,		1, TIFF_ASCII },
-	{ TIFFTAG_WHITEPOINT,		1, TIFF_RATIONAL },
+	{ TIFFTAG_WHITEPOINT,		(uint16) -1,TIFF_RATIONAL },
 	{ TIFFTAG_PRIMARYCHROMATICITIES,(uint16) -1,TIFF_RATIONAL },
 	{ TIFFTAG_HALFTONEHINTS,	2, TIFF_SHORT },
 	{ TIFFTAG_BADFAXLINES,		1, TIFF_LONG },
@@ -626,7 +678,7 @@ tiffcp(TIFF* in, TIFF* out)
 	}
 	cpOtherTags(in, out);
 
-	if (geofile)
+	if (geofile || proj4_string )
             InstallGeoTIFF(out);
         else
             CopyGeoTIFF(in,out);
@@ -689,7 +741,7 @@ DECLAREcpFunc(cpContig2ContigByRow_8_to_4)
 
     printf( "Downsample\n" );
 
-    (void) imagewidth; (void) spp;
+    (void) spp;
     for (row = 0; row < imagelength; row++) {
         int i_in, i_out_byte;
             
@@ -697,7 +749,7 @@ DECLAREcpFunc(cpContig2ContigByRow_8_to_4)
             goto done;
 
         for( i_in = 0, i_out_byte = 0;
-             i_in < imagewidth;
+             (unsigned)i_in < imagewidth;
              i_in += 2, i_out_byte++ )
         {
             buf_out[i_out_byte] =
